"""Module for interacting with a JVC Projector."""

from __future__ import annotations
from typing import Mapping
from . import command
from .command import JvcCommand, JvcCommandHelpers
from .connection import resolve
from .device import JvcDevice
from .error import JvcProjectorConnectError, JvcProjectorError
from . import const
import logging

DEFAULT_PORT = 20554
DEFAULT_TIMEOUT = 15.0

_LOGGER = logging.getLogger(__name__)


class JvcProjector:
    """Class for interacting with a JVC Projector."""

    def __init__(
        self,
        host: str,
        *,
        port: int = DEFAULT_PORT,
        timeout: float = DEFAULT_TIMEOUT,
        password: str | None = None,
    ) -> None:
        """Initialize class."""
        self._host = host
        self._port = port
        self._timeout = timeout
        self._password = password

        self._device: JvcDevice | None = None
        self._ip: str = ""
        self._model: str = ""
        self._mac: str = ""
        self._version: str = ""
        self._dict: dict[str, str] = {}

    @property
    def ip(self) -> str:
        """Returns ip."""
        if not self._ip:
            raise JvcProjectorError("ip not initialized")
        return self._ip

    @property
    def host(self) -> str:
        """Returns host."""
        return self._host

    @property
    def port(self) -> int:
        """Returns port."""
        return self._port

    @property
    def model(self) -> str:
        """Returns model name."""
        if not self._mac:
            raise JvcProjectorError("model not initialized")
        return self.process_model_code(self._model)

    @property
    def mac(self) -> str:
        """Returns mac address."""
        if not self._mac:
            raise JvcProjectorError("mac address not initialized")
        return self._mac

    @property
    def version(self) -> str:
        """Get device software version."""
        if not self._version:
            raise JvcProjectorError("version address not initialized")
        return self.process_version(self._version)

    async def is_on(self) -> bool:
        """Return if device is fully on and ready."""
        return await self.get_power() == const.ON

    async def connect(self, get_info: bool = False) -> None:
        """Connect to device."""
        if self._device:
            return

        if not self._ip:
            self._ip = await resolve(self._host)

        self._device = JvcDevice(self._ip, self._port, self._timeout, self._password)

        if not await self.test():
            raise JvcProjectorConnectError("Failed to verify connection")

        if get_info:
            await self.get_info()

    async def disconnect(self) -> None:
        """Disconnect from device."""
        if self._device:
            await self._device.disconnect()
            self._device = None

    async def get_info(self) -> dict[str, str]:
        """Get device info.

        Returns:
            dict[str, str]: Dictionary containing model and MAC address

        Raises:
            JvcProjectorError: If device is not connected or MAC address is unavailable

        """
        if not self._device:
            raise JvcProjectorError("Must call connect before getting info")

        # Create base commands that work regardless of power state
        model = JvcCommand(const.CMD_MODEL, True)
        mac = JvcCommand(const.CMD_LAN_SETUP_MAC_ADDRESS, True)
        commands = [model, mac]
        _LOGGER.debug("Getting info with commands %s", commands)

        # Only add version command if projector is on
        is_on = await self.is_on()
        if is_on:
            _LOGGER.debug("Projector is on, adding version command")
            version = JvcCommand(const.CMD_VERSION, True)
            commands.append(version)

        # Send commands
        await self._send(commands)
        _LOGGER.debug("Got responses %s", [cmd.response for cmd in commands])

        # Validate
        if mac.response is None:
            raise JvcProjectorError("MAC address not available")

        # responses
        self._model = model.response or "(unknown)"
        self._mac = mac.response  # None checked above
        self._version = version.response if is_on and version.response else "(unknown)"

        return {const.KEY_MODEL: self._model, const.KEY_MAC: self._mac}

    async def get_state(self) -> Mapping[str, str | None]:
        """Get device state."""
        if not self._device:
            raise JvcProjectorError("Must call connect before getting state")

        # Add static values
        self._dict[const.KEY_MODEL] = self.process_model_code(self._model)
        self._dict[const.KEY_VERSION] = self.process_version(self._version)
        self._dict[const.KEY_MAC] = self.process_mac(self._mac)

        async def send_and_update(commands: dict[str, str]) -> None:
            """Send commands and update the dictionary."""
            # send commands with is_ref set to true
            cmd_vals = [JvcCommand(cmd, True) for cmd in commands.values()]
            res = await self._send(cmd_vals)
            # discard the command values and zip the keys with the responses
            for (key, _), value in zip(commands.items(), res):
                # Only store non-None values
                if value is not None:
                    self._dict[key] = value

        # Always get power state
        await send_and_update({const.KEY_POWER: const.CMD_POWER})

        # If power is on, get additional states
        if self._dict.get(const.KEY_POWER) == const.ON:
            await send_and_update(
                {
                    const.KEY_INPUT: const.CMD_INPUT,
                    const.KEY_SOURCE: const.CMD_SOURCE,
                    const.KEY_PICTURE_MODE: const.CMD_PICTURE_MODE,
                    const.KEY_LOW_LATENCY: const.CMD_PICTURE_MODE_LOW_LATENCY,
                    const.KEY_LASER_POWER: const.CMD_PICTURE_MODE_LASER_POWER,
                    const.KEY_LASER_TIME: const.CMD_FUNCTION_LASER_TIME,
                    const.KEY_ANAMORPHIC: const.CMD_INSTALLATION_ANAMORPHIC,
                    const.KEY_INSTALLATION_MODE: const.CMD_INSTALLATION_MODE,
                }
            )

            # Check if there's a signal before getting signal-dependent states
            if self._dict.get(const.KEY_SOURCE) == const.SIGNAL:
                await send_and_update(
                    {
                        const.KEY_HDR: const.CMD_FUNCTION_HDR,
                        const.KEY_HDMI_INPUT_LEVEL: const.CMD_INPUT_SIGNAL_HDMI_INPUT_LEVEL,
                        const.KEY_HDMI_COLOR_SPACE: const.CMD_INPUT_SIGNAL_HDMI_COLOR_SPACE,
                        const.KEY_COLOR_PROFILE: const.CMD_PICTURE_MODE_COLOR_PROFILE,
                        const.KEY_GRAPHICS_MODE: const.CMD_PICTURE_MODE_GRAPHICS_MODE,
                        const.KEY_COLOR_SPACE: const.CMD_FUNCTION_COLOR_SPACE,
                        const.KEY_RESOLUTION: const.CMD_FUNCTION_SOURCE,
                    }
                )

            # NX9 and NZ model specific commands
            if (
                "NZ" in self._dict[const.KEY_MODEL]
                or "NX9" in self._dict[const.KEY_MODEL]
            ):
                await send_and_update(
                    {
                        const.KEY_ESHIFT: const.CMD_PICTURE_MODE_8K_ESHIFT,
                        const.KEY_CLEAR_MOTION_DRIVE: const.CMD_PICTURE_MODE_CLEAR_MOTION_DRIVE,
                        const.KEY_MOTION_ENHANCE: const.CMD_PICTURE_MODE_MOTION_ENHANCE,
                        const.KEY_LASER_VALUE: const.CMD_PICTURE_MODE_LASER_VALUE,
                        const.KEY_LASER_DIMMING: const.CMD_LASER_DIMMING,
                    }
                )

            # HDR-specific commands
            if (
                self._dict.get("hdr")
                not in [const.HDR_CONTENT_NONE, const.HDR_CONTENT_SDR]
            ) and self._dict.get(const.KEY_SOURCE) == const.SIGNAL:
                await send_and_update(
                    {
                        const.KEY_HDR_PROCESSING: const.CMD_PICTURE_MODE_HDR_PROCESSING,
                        const.KEY_HDR_CONTENT_TYPE: const.CMD_PICTURE_MODE_HDR_CONTENT_TYPE,
                    }
                )

        return self._dict

    def process_mac(self, mac: str) -> str:
        """Process mac address."""
        # skip every 2 characters and join with :
        return ":".join(mac[i : i + 2] for i in range(0, len(mac), 2))

    def process_model_code(self, model: str) -> str:
        """Process model code."""
        return const.MODEL_MAP.get(model[-4:], f"{model}: Unknown")

    def process_version(self, version: str) -> str:
        """Process version string."""

        if version == "(unknown)":
            return version

        version = version.removesuffix("PJ")
        version = version.zfill(4)

        # Extract major, minor, and patch versions
        major = str(
            int(version[0:2])
        )  # Remove leading zero and convert to int then back to str
        minor = str(int(version[2]))
        patch = str(int(version[3]))

        return f"{major}.{minor}.{patch}"

    async def get_version(self) -> str | None:
        """Get device software version."""
        return await self.ref(const.CMD_VERSION)

    async def get_power(self) -> str | None:
        """Get power state."""
        return await self.ref(const.CMD_POWER)

    async def get_input(self) -> str | None:
        """Get current input."""
        return await self.ref(const.CMD_INPUT)

    async def get_signal(self) -> str | None:
        """Get if has signal."""
        return await self.ref(const.CMD_SOURCE)

    async def test(self) -> bool:
        """Run test command."""
        cmd = JvcCommand(f"{command.TEST}")
        await self._send([cmd])
        return cmd.ack

    async def power_on(self) -> None:
        """Run power on command."""
        await self.op(f"{const.CMD_POWER}1")

    async def power_off(self) -> None:
        """Run power off command."""
        await self.op(f"{const.CMD_POWER}0")

    async def remote(self, code: str) -> None:
        """Run remote code command."""
        await self.op(f"{const.CMD_REMOTE}{code}")

    async def op(self, code: str) -> None:
        """Send operation code."""
        await self._send([JvcCommand(code, False)])

    async def ref(self, code: str) -> str | None:
        """Send reference code."""
        return (await self._send([JvcCommand(code, True)]))[0]

    async def _send(self, cmds: list[JvcCommand]) -> list[str | None]:
        """Send command to device."""
        if self._device is None:
            raise JvcProjectorError("Must call connect before sending commands")

        await self._device.send(cmds)

        return [cmd.response for cmd in cmds]

    @staticmethod
    def _invert_dict(d: dict[str, str]) -> dict[str, str]:
        return {v: k for k, v in d.items()}

    async def send_command(self, cmd: str, val: str) -> None:
        """Send a command to the projector using well-known names like "power". Intended to be human readable commands."""
        # normalize the command and value
        cmd = cmd.lower()
        val = val.lower()

        # get all valid commands and the PJ codes
        valid_command_map = JvcCommandHelpers.get_available_commands()

        # ensure the command is valid
        if cmd not in valid_command_map:
            raise ValueError(f"Unknown command: {cmd}")

        # get the PJ code and values
        cmd_values = valid_command_map[cmd]["values"]
        raw_cmd = valid_command_map[cmd]["command"]

        # ensure the value is valid
        try:
            if val not in cmd_values:
                raise ValueError(f"Invalid value for {cmd}: {val}")
        except KeyError as exc:
            raise ValueError(f"Unknown command: {cmd}") from exc

        # get the PJ values like high = 1
        command_info = JvcCommand.command_map[raw_cmd]

        # transform human values to PJ values
        if command_info["values"] == "callable":
            # For callable formatters, we'll just pass the value as is
            value = val
        elif "inverse" in command_info:
            if val not in command_info["inverse"]:
                raise ValueError(f"Invalid value for {cmd}: {val}")
            value = command_info["inverse"][val]
        else:
            raise ValueError(f"Unsupported command type for {cmd}")

        await self.op(f"{raw_cmd}{value}")



"""Custom errors for a JVC Projector."""


class JvcProjectorError(Exception):
    """Projector Error."""


class JvcProjectorConnectError(JvcProjectorError):
    """Projector Connect Timeout."""


class JvcProjectorCommandError(JvcProjectorError):
    """Projector Command Error."""


class JvcProjectorAuthError(JvcProjectorError):
    """Projector Password Invalid Error."""



"""Module for representing a JVC Projector command."""

from __future__ import annotations

from collections.abc import Callable
import logging
import re
from typing import Final, Any
import math
from . import const

_LOGGER = logging.getLogger(__name__)

PJOK: Final = b"PJ_OK"
PJNG: Final = b"PJ_NG"
PJREQ: Final = b"PJREQ"
PJACK: Final = b"PJACK"
PJNAK: Final = b"PJNAK"

UNIT_ID: Final = b"\x89\x01"
HEAD_OP: Final = b"!" + UNIT_ID
HEAD_REF: Final = b"?" + UNIT_ID
HEAD_RES: Final = b"@" + UNIT_ID
HEAD_ACK: Final = b"\x06" + UNIT_ID
HEAD_LEN: Final = 1 + len(UNIT_ID)
END: Final = b"\n"

TEST: Final = "\0\0"

AUTH_SALT: Final = "JVCKWPJ"


class JvcCommand:
    """Class for representing a JVC Projector command."""

    def __init__(self, code: str, is_ref=False):
        """Initialize class."""
        self.code = code
        self.is_ref = is_ref
        self.ack = False
        self._response: str | None = None

    @property
    def response(self) -> str | None:
        """Return command response."""
        if not self.is_ref or self._response is None:
            return None

        res = self.code + self._response
        val: str = res[len(self.code) :]

        for pat, fmt in self.formatters.items():
            m = re.search(r"^" + pat + r"$", res)
            if m:
                if isinstance(fmt, list):
                    try:
                        index = int(m[1], 16)
                        if 0 <= index < len(fmt):
                            return fmt[index]
                        return val  # noqa: TRY300
                    except ValueError:
                        msg = "response '%s' not int for cmd '%s'"
                        _LOGGER.warning(msg, val, self.code)
                elif isinstance(fmt, dict):
                    try:
                        return fmt[m[1]]
                    except KeyError:
                        msg = "response '%s' not mapped for cmd '%s'"
                        _LOGGER.warning(msg, val, self.code)
                elif callable(fmt):
                    try:
                        return fmt(m)
                    except Exception as e:  # noqa: BLE001
                        msg = "response format failed with %s for '%s (%s)'"
                        _LOGGER.warning(msg, e, self.code, val)
                break

        return val

    @response.setter
    def response(self, data: str) -> None:
        """Set command response."""
        self._response = data

    @property
    def is_power(self) -> bool:
        """Return if command is a power command."""
        return self.code.startswith("PW")

    @staticmethod
    def _build_command_map(
        formatters: dict[str, list | dict | Callable],
    ) -> dict[str, dict[str, Any]]:
        """Use the Formatters object in JvcCommand to build a command map to reduce code duplication."""
        command_map = {}
        for pattern, formatter in formatters.items():
            opcode = pattern.split("(")[0]  # Extract opcode from the f-string
            if isinstance(formatter, dict):
                command_map[opcode] = {
                    "values": formatter,
                    "inverse": {v: k for k, v in formatter.items()},
                }
            elif isinstance(formatter, list):
                command_map[opcode] = {
                    "values": {
                        str(i): val
                        for i, val in enumerate(formatter)
                        if val is not None
                    },
                    "inverse": {
                        val: str(i)
                        for i, val in enumerate(formatter)
                        if val is not None
                    },
                }
            elif callable(formatter):
                command_map[opcode] = {"values": {"callable": "callable"}}
            else:
                command_map[opcode] = {"values": formatter}
        return command_map

    formatters: dict[str, list | dict | Callable] = {
        f"{const.CMD_POWER}(.)": const.VAL_POWER,
        f"{const.CMD_INPUT}(.)": const.VAL_INPUT,
        f"{const.CMD_SOURCE}(.)": const.VAL_SOURCE,
        f"{const.CMD_MODEL}(.+)": lambda r: re.sub(r"-+", "-", r[1].replace(" ", "-")),
        f"{const.CMD_PICTURE_MODE}(..)": const.VAL_PICTURE_MODE,
        f"{const.CMD_PICTURE_MODE_INTELLIGENT_APERTURE}(.)": const.VAL_INTELLIGENT_APERTURE,
        f"{const.CMD_LASER_DIMMING}(.)": const.VAL_LASER_DIMMING,
        f"{const.CMD_PICTURE_MODE_COLOR_PROFILE}(..)": const.VAL_COLOR_PROFILE,
        f"{const.CMD_PICTURE_MODE_COLOR_TEMP}(..)": const.VAL_COLOR_TEMP,
        f"{const.CMD_PICTURE_MODE_COLOR_CORRECTION}(..)": const.VAL_COLOR_CORRECTION,
        f"{const.CMD_PICTURE_MODE_GAMMA_TABLE}(..)": const.VAL_GAMMA_TABLE,
        f"{const.CMD_PICTURE_MODE_COLOR_MANAGEMENT}(.)": const.VAL_TOGGLE,
        f"{const.CMD_PICTURE_MODE_LOW_LATENCY}(.)": const.VAL_TOGGLE,
        f"{const.CMD_PICTURE_MODE_8K_ESHIFT}(.)": const.VAL_TOGGLE,
        f"{const.CMD_PICTURE_MODE_CLEAR_MOTION_DRIVE}(.)": const.VAL_CLEAR_MOTION_DRIVE,
        f"{const.CMD_PICTURE_MODE_LASER_VALUE}(.+)": lambda r: math.floor(
            ((int(r[1], 16) - 109) / 1.1) + 0.5
        ),
        f"{const.CMD_PICTURE_MODE_MOTION_ENHANCE}(.)": const.VAL_MOTION_ENHANCE,
        f"{const.CMD_PICTURE_MODE_LASER_POWER}(.)": const.VAL_LASER_POWER,
        f"{const.CMD_PICTURE_MODE_GRAPHICS_MODE}(.)": const.VAL_GRAPHICS_MODE,
        f"{const.CMD_INPUT_SIGNAL_HDMI_INPUT_LEVEL}(.)": const.VAL_HDMI_INPUT_LEVEL,
        f"{const.CMD_INPUT_SIGNAL_HDMI_COLOR_SPACE}(.)": const.VAL_HDMI_COLOR_SPACE,
        f"{const.CMD_INPUT_SIGNAL_HDMI_2D_3D}(.)": const.VAL_HDMI_2D_3D,
        f"{const.CMD_INPUT_SIGNAL_ASPECT}(.)": const.VAL_ASPECT,
        f"{const.CMD_INPUT_SIGNAL_MASK}(.)": const.VAL_MASK,
        f"{const.CMD_INSTALLATION_MODE}(.)": const.VAL_INSTALLATION_MODE,
        f"{const.CMD_INSTALLATION_LENS_CONTROL}(.)": const.VAL_LENS_CONTROL,
        f"{const.CMD_INSTALLATION_LENS_IMAGE_PATTERN}(.)": const.VAL_TOGGLE,
        f"{const.CMD_INSTALLATION_LENS_LOCK}(.)": const.VAL_TOGGLE,
        f"{const.CMD_INSTALLATION_SCREEN_ADJUST}(.)": const.VAL_TOGGLE,
        f"{const.CMD_INSTALLATION_STYLE}(.)": const.VAL_INSTALLATION_STYLE,
        f"{const.CMD_INSTALLATION_ANAMORPHIC}(.)": const.VAL_ANAMORPHIC,
        f"{const.CMD_DISPLAY_BACK_COLOR}(.)": const.VAL_BACK_COLOR,
        f"{const.CMD_DISPLAY_MENU_POSITION}(.)": const.VAL_MENU_POSITION,
        f"{const.CMD_DISPLAY_SOURCE_DISPLAY}(.)": const.VAL_TOGGLE,
        f"{const.CMD_DISPLAY_LOGO}(.)": const.VAL_TOGGLE,
        f"{const.CMD_FUNCTION_TRIGGER}(.)": const.VAL_TRIGGER,
        f"{const.CMD_FUNCTION_OFF_TIMER}(.)": const.VAL_OFF_TIMER,
        f"{const.CMD_FUNCTION_ECO_MODE}(.)": const.VAL_TOGGLE,
        f"{const.CMD_FUNCTION_CONTROL4}(.)": const.VAL_TOGGLE,
        f"{const.CMD_FUNCTION_INPUT}(.)": const.VAL_FUNCTION_INPUT,
        f"{const.CMD_FUNCTION_SOURCE}(..)": const.VAL_FUNCTION_SOURCE,
        f"{const.CMD_FUNCTION_DEEP_COLOR}(.)": const.VAL_DEEP_COLOR,
        f"{const.CMD_FUNCTION_COLOR_SPACE}(.)": const.VAL_COLOR_SPACE,
        f"{const.CMD_FUNCTION_COLORIMETRY}(.)": const.VAL_COLORIMETRY,
        f"{const.CMD_FUNCTION_HDR}(.)": const.VAL_HDR,
        f"{const.CMD_FUNCTION_LASER_TIME}(.+)": lambda r: int(r[1], 16),
        f"{const.CMD_PICTURE_MODE_HDR_LEVEL}(.)": const.VAL_HDR_LEVEL,
        f"{const.CMD_PICTURE_MODE_HDR_PROCESSING}(.)": const.VAL_HDR_PROCESSING,
        f"{const.CMD_PICTURE_MODE_HDR_CONTENT_TYPE}(.)": const.VAL_HDR_CONTENT_TYPE,
        f"{const.CMD_PICTURE_MODE_THEATER_OPTIMIZER}(.)": const.VAL_TOGGLE,
        f"{const.CMD_PICTURE_MODE_THEATER_OPTIMIZER_LEVEL}(.)": const.VAL_THEATER_OPTIMIZER_LEVEL,
        f"{const.CMD_PICTURE_MODE_THEATER_OPTIMIZER_PROCESSING}(.)": const.VAL_THEATER_OPTIMIZER_PROCESSING,
        f"{const.CMD_LAN_SETUP_DHCP}(.)": const.VAL_TOGGLE,
        f"{const.CMD_LAN_SETUP_MAC_ADDRESS}(.+)": lambda r: re.sub(
            r"-+", "-", r[1].replace(" ", "-")
        ),
        f"{const.CMD_LAN_SETUP_IP_ADDRESS}(..)(..)(..)(..)": lambda r: f"{int(r[1], 16)}.{int(r[2], 16)}.{int(r[3], 16)}.{int(r[4], 16)}",
    }

    command_map: dict[str, dict[str, Any]] = _build_command_map(formatters)

    @staticmethod
    def get_available_commands() -> dict[str, str]:
        """Return a map of commands to their allowed values/responses based on the available command_map."""

        human_commands: dict[str, str] = {}
        for cmd, values in JvcCommandHelpers.get_available_commands().items():
            human_commands[cmd] = values["values"]
        return human_commands


class JvcCommandHelpers:
    """Helper methods for JVC Command handling and documentation."""

    # Define override maps for commands where input values differ from return values
    COMMAND_VALUE_OVERRIDES = {
        "power": [
            "on",
            "standby",
        ],  # Can only send standby/on, even though we might receive other states
        "source": None,  # Read-only, can't send values
        "laser_time": None,  # Read-only, can't send values
    }

    @staticmethod
    def format_value_name(value: str) -> str:
        """Format a command value into a human-readable string."""
        # Replace underscores with spaces
        formatted = value.replace("_", " ")

        # make sure all words are lowercase
        return " ".join(word.lower() for word in formatted.split())

    @staticmethod
    def format_command_name(command: str) -> str:
        """Format a command name into a human-readable string."""
        # Remove KEY_ prefix if present
        command = re.sub(r"^KEY_", "", command)
        # Replace underscores with spaces
        formatted = command.replace("_", " ")
        # Capitalize first letter of each word
        return " ".join(word.capitalize() for word in formatted.split())

    @staticmethod
    def format_value_list(values: list[str] | dict[str, str]) -> list[str]:
        """Format a list or dict of values into human-readable strings."""
        if isinstance(values, dict):
            # For dictionaries, we only want the values, not the keys
            value_list = list(values.values())
        else:
            value_list = values

        return [JvcCommandHelpers.format_value_name(str(val)) for val in value_list]

    @classmethod
    def get_command_values(
        cls, command: str, values: list[str] | dict[str, str]
    ) -> list[str] | None:
        """Get the appropriate command values, taking into account any overrides.

        Args:
            command: The command key
            values: The default values from the command map

        Returns:
            List of valid input values for the command or None if read-only

        """
        # Check if we have an override for this command
        if command in cls.COMMAND_VALUE_OVERRIDES:
            return cls.COMMAND_VALUE_OVERRIDES[command]

        # No override, process normal values
        return cls.format_value_list(values)

    @classmethod
    def get_available_commands(cls) -> dict[str, dict[str, Any]]:
        """Return a detailed map of commands and their allowed values/responses.

        Returns:
            Dict with structure:
            {
                "command_name": {
                    "name": "Human Readable Name",
                    "values": ["Value 1", "Value 2"],
                    "description": "Command description",
                    "type": "enum|range|read_only",
                    "command": "PJ_COMMAND"  # The raw projector command
                }
            }

        """
        human_commands: dict[str, dict[str, Any]] = {}

        # Command category descriptions
        categories = {
            "POWER": "Controls projector power state",
            "INPUT": "Selects input source",
            "PICTURE_MODE": "Controls picture mode settings",
            "INSTALLATION": "Controls installation and setup options",
            "HDR": "Controls HDR-related settings",
            "LASER": "Controls laser and brightness settings",
            "COLOR": "Controls color-related settings",
            "MOTION": "Controls motion-related settings",
        }

        for human_cmd, pj_cmd in const.KEY_MAP_TO_COMMAND.items():
            command_data = JvcCommand.command_map.get(pj_cmd, {})

            # Skip if no command data found
            if not command_data:
                continue

            # Determine command category
            category = next(
                (cat for cat in categories if cat in human_cmd.upper()),
                "GENERAL",
            )

            values = command_data.get("values", {})

            command_info = {
                "name": cls.format_command_name(human_cmd),
                "category": category,
                "description": categories.get(category, "General setting"),
                "command": pj_cmd,
            }

            # Handle different value types
            if values == "callable":
                command_info["values"] = '["0-100"]'
                command_info["type"] = "range"
            elif isinstance(values, (list, dict)):
                command_values = cls.get_command_values(human_cmd, values)
                if command_values is None:
                    command_info["values"] = ""
                    command_info["type"] = "read_only"
                else:
                    command_info["values"] = ", ".join(
                        f'"{val}"' for val in command_values
                    )
                    command_info["type"] = "enum"
            else:
                command_info["values"] = ""
                command_info["type"] = "unknown"

            human_commands[human_cmd] = command_info

        return human_commands

    @classmethod
    def get_command_help(cls, command: str) -> str:
        """Generate a human-readable help string for a specific command.

        Args:
            command: The command key to get help for

        Returns:
            A formatted help string with command details

        """
        commands = cls.get_available_commands()
        if command not in commands:
            return f"Unknown command: {command}"

        cmd_info = commands[command]

        help_text = [
            f"Command: {cmd_info['name']}",
            f"Category: {cmd_info['category']}",
            f"Description: {cmd_info['description']}",
        ]

        if cmd_info["type"] == "read_only":
            help_text.append("Type: Read Only")
        elif cmd_info["type"] == "range":
            help_text.append("Accepts values: 0-100")
        elif cmd_info["values"]:
            help_text.append("Available values:")
            help_text.extend(f"  - {value}" for value in cmd_info["values"])

        return "\n".join(help_text)



"""Functional tests."""

import asyncio
import logging
import os
import pytest
from jvcprojector.projector import JvcProjector
from jvcprojector import const

logging.basicConfig(level=logging.WARNING)
_LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
@pytest.mark.skipif(
    os.environ.get("RUN_INTEGRATION") != "true", reason="Set RUN_INTEGRATION to true"
)
async def test_connect():
    """Test connect succeeds."""
    ip = os.getenv("JVC_IP")
    password = os.getenv("JVC_PASSWORD")
    if not ip or not password:
        _LOGGER.error("Set JVC_IP and JVC_PASSWORD environment variables")
        pytest.fail("Set JVC_IP and JVC_PASSWORD environment variables")

    jp = JvcProjector(ip, password=password)
    await jp.connect()

    _LOGGER.info("Projector info:")
    _LOGGER.info(await jp.get_info())

    if not await jp.is_on():
        await jp.power_on()
        _LOGGER.info("Waiting for projector to warmup...")
        while not await jp.is_on():
            await asyncio.sleep(3)

    await jp.send_command(const.KEY_LASER_POWER, const.MEDIUM)
    _LOGGER.info("Current state:")
    state = await jp.get_state()
    for key, value in state.items():
        _LOGGER.info("%s: %s", key, value)

    await jp.disconnect()



"""Tests for projector module."""

from unittest.mock import AsyncMock

from unittest.mock import patch
import pytest

from jvcprojector import const
from jvcprojector.error import JvcProjectorError
from jvcprojector.projector import JvcProjector

from . import HOST, IP, MAC, MODEL, PORT, VERSION


@pytest.mark.asyncio
async def test_init(dev: AsyncMock):
    """Test init succeeds."""
    p = JvcProjector(IP, port=PORT)
    assert p.host == IP
    assert p.port == PORT
    with pytest.raises(JvcProjectorError):
        assert p.ip
    with pytest.raises(JvcProjectorError):
        assert p.model
    with pytest.raises(JvcProjectorError):
        assert p.mac


@pytest.mark.asyncio
async def test_connect(dev: AsyncMock):
    """Test connect succeeds."""
    p = JvcProjector(IP, port=PORT)
    await p.connect()
    assert p.ip == IP
    await p.disconnect()
    assert dev.disconnect.call_count == 1


@pytest.mark.asyncio
async def test_connect_host(dev: AsyncMock):
    """Test connect succeeds."""
    p = JvcProjector(HOST, port=PORT)
    await p.connect()
    assert p.ip == IP
    await p.disconnect()
    assert dev.disconnect.call_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("dev", [{const.CMD_MODEL: None}], indirect=True)
async def test_unknown_model(dev: AsyncMock):
    """Test projector with unknown model succeeds."""
    p = JvcProjector(IP)
    await p.connect()
    await p.get_info()
    assert p.mac == MAC
    assert p.model == "(unknown): Unknown"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "dev", [{const.CMD_LAN_SETUP_MAC_ADDRESS: None}], indirect=True
)
async def test_unknown_mac(dev: AsyncMock):
    """Test projector with unknown mac uses model succeeds."""
    p = JvcProjector(IP)
    await p.connect()
    with pytest.raises(JvcProjectorError):
        assert p.mac == ""


@pytest.mark.asyncio
async def test_get_info(dev: AsyncMock):
    """Test get_info succeeds."""
    p = JvcProjector(IP)
    await p.connect()
    assert await p.get_info() == {"model": MODEL, "mac": MAC}
    # make sure model code is right
    assert p.model in const.MODEL_MAP.values()
    # make sure version is right
    assert p.version == "3.0.0"


@pytest.mark.asyncio
async def test_get_state(dev: AsyncMock):
    """Test get_state succeeds."""
    p = JvcProjector(IP)
    await p.connect()
    expected_dict = {
        "power": const.ON,
    }
    assert set(expected_dict.items()).issubset((await p.get_state()).items())


def test_version():
    """Test version processing."""

    p = JvcProjector(IP)
    assert p.process_version(VERSION) == "3.0.0"


def test_model():
    """Test model processing."""
    model = "B5A1"
    p = JvcProjector(IP)
    assert p.process_model_code(model) == "NZ9"


@pytest.mark.asyncio
async def test_send_command_success(dev: AsyncMock):
    """Test send_command succeeds."""
    p = JvcProjector(IP, port=PORT)
    await p.connect()

    # Mock the _build_command_map method
    with patch("jvcprojector.projector.JvcCommand._build_command_map") as mock_build:
        mock_build.return_value = {
            const.CMD_PICTURE_MODE_LASER_POWER: {
                "values": {"0": "low", "1": "high", "2": "medium"},
                "inverse": {"low": "0", "high": "1", "medium": "2"},
            }
        }

    # Patch the op method to avoid actual sending
    with patch.object(p, "op") as mock_op:
        await p.send_command(const.KEY_LASER_POWER, const.HIGH)

        # Assert that op was called once with the correct command
        mock_op.assert_called_once_with(f"{const.CMD_PICTURE_MODE_LASER_POWER}1")

    # We don't need to assert anything about dev.send here, as we're not directly testing it


@pytest.mark.asyncio
async def test_send_command_invalid_value(dev: AsyncMock):
    """Test send_command fails with invalid value."""
    p = JvcProjector(IP, port=PORT)
    await p.connect()

    # Mock the _build_command_map method
    with patch("jvcprojector.projector.JvcCommand._build_command_map") as mock_build:
        mock_build.return_value = {
            const.CMD_PICTURE_MODE_LASER_POWER: {
                "values": {"0": "low", "1": "high", "2": "medium"},
                "inverse": {"low": "0", "high": "1", "medium": "2"},
            }
        }

    with pytest.raises(
        ValueError,
        match=f"Invalid value for {const.KEY_LASER_POWER}: invalid_value",
    ):
        await p.send_command(const.KEY_LASER_POWER, "invalid_value")


@pytest.mark.asyncio
async def test_send_command_unknown_command(dev: AsyncMock):
    """Test send_command fails with unknown command."""
    p = JvcProjector(IP, port=PORT)
    await p.connect()

    # Mock the _build_command_map method
    with patch("jvcprojector.projector.JvcCommand._build_command_map") as mock_build:
        mock_build.return_value = {}

        with pytest.raises(ValueError, match="Unknown command: unknown_command"):
            await p.send_command("unknown_command", "value")



"""Module for representing a JVC Projector device."""

from __future__ import annotations

import asyncio
from hashlib import sha256
import logging
import struct
from time import time

from . import const
from .command import (
    AUTH_SALT,
    END,
    HEAD_ACK,
    HEAD_LEN,
    HEAD_OP,
    HEAD_REF,
    HEAD_RES,
    PJACK,
    PJNAK,
    PJNG,
    PJOK,
    PJREQ,
    JvcCommand,
)
from .connection import JvcConnection
from .error import (
    JvcProjectorAuthError,
    JvcProjectorCommandError,
    JvcProjectorConnectError,
)

KEEPALIVE_TTL = 2

_LOGGER = logging.getLogger(__name__)


class JvcDevice:
    """Class for representing a JVC Projector device."""

    def __init__(
        self, ip: str, port: int, timeout: float, password: str | None = None
    ) -> None:
        """Initialize class."""
        self._conn = JvcConnection(ip, port, timeout)

        self._auth = b""
        if password:
            self._auth = struct.pack(f"{max(10, len(password))}s", password.encode())

        self._lock = asyncio.Lock()
        self._keepalive: asyncio.Task | None = None
        self._last: float = 0.0

    async def send(self, cmds: list[JvcCommand]) -> None:
        """Send commands to device."""
        async with self._lock:
            # Treat status refreshes with special handling
            # if any cmd is a ref, and the first cmd is a power cmd
            is_refresh = (
                any(cmd.is_ref for cmd in cmds) and len(cmds) > 1 and cmds[0].is_power
            )
            _LOGGER.debug("Sending %s", "refresh" if is_refresh else "commands")

            # Connection keepalive window for fast command repeats
            keepalive = True

            # Connection keepalive window for fast command repeats
            if self._keepalive:
                self._keepalive.cancel()
                self._keepalive = None
            elif is_refresh:
                # Don't extend window below if this was a refresh
                keepalive = False

            try:
                if not self._conn.is_connected():
                    await self._connect()

                cmd = None

                for cmd in cmds:
                    await self._send(cmd)
                    # Throttle since some projectors dont like back to back commands
                    # tested that as low as 0.1 is okay
                    # however, it IS possible to lock up even NZ models
                    await asyncio.sleep(0.2)
                    # If device is not powered on, skip remaining commands
                    # mac cannot be is_refresh here and first command must be pw if its a refresh
                    if is_refresh and cmds[0].response != const.ON:
                        _LOGGER.debug(
                            "Skipping commands due to power off which are %s",
                            [c.code for c in cmds],
                        )
                        break
            except Exception:
                keepalive = False
                raise
            finally:
                # Delay disconnect to keep connection alive.
                if keepalive and cmd and cmd.ack:
                    self._keepalive = asyncio.create_task(
                        self._disconnect(KEEPALIVE_TTL)
                    )
                else:
                    await self._disconnect()

    async def _connect(self) -> None:
        """Connect to device."""
        assert not self._conn.is_connected()

        elapsed = time() - self._last
        if elapsed < 0.75:
            await asyncio.sleep(0.75 - elapsed)

        retries = 0
        while retries < 10:
            try:
                _LOGGER.debug("Connecting to %s", self._conn.ip)
                await self._conn.connect()
            except ConnectionRefusedError:
                retries += 1
                if retries == 5:
                    _LOGGER.warning("Retrying refused connection")
                else:
                    _LOGGER.debug("Retrying refused connection")
                await asyncio.sleep(0.2 * retries)
                continue
            except (asyncio.TimeoutError, ConnectionError) as err:
                raise JvcProjectorConnectError from err

            try:
                data = await self._conn.read(len(PJOK))
            except asyncio.TimeoutError as err:
                raise JvcProjectorConnectError("Handshake init timeout") from err

            _LOGGER.debug("Handshake received %s", data)

            if data == PJNG:
                _LOGGER.warning("Handshake retrying on busy")
                retries += 1
                await asyncio.sleep(0.25 * retries)
                continue

            if data != PJOK:
                raise JvcProjectorCommandError("Handshake init invalid")

            break
        else:
            raise JvcProjectorConnectError("Retries exceeded")

        _LOGGER.debug("Handshake sending '%s'", PJREQ.decode())
        await self._conn.write(PJREQ + (b"_" + self._auth if self._auth else b""))

        try:
            data = await self._conn.read(len(PJACK))
            _LOGGER.debug("Handshake received %s", data)

            if data == PJNAK:
                _LOGGER.debug("Standard auth failed, trying SHA256 auth")
                auth = (
                    sha256(f"{self._auth.decode()}{AUTH_SALT}".encode())
                    .hexdigest()
                    .encode()
                )
                await self._conn.write(PJREQ + b"_" + auth)
                data = await self._conn.read(len(PJACK))
                if data == PJACK:
                    self._auth = auth

            if data == PJNAK:
                raise JvcProjectorAuthError

            if data != PJACK:
                raise JvcProjectorCommandError("Handshake ack invalid")

        except asyncio.TimeoutError as err:
            raise JvcProjectorConnectError("Handshake ack timeout") from err

        self._last = time()

    async def _send(self, cmd: JvcCommand) -> None:
        """Send command to device."""
        assert self._conn.is_connected()
        assert len(cmd.code) >= 2

        code = cmd.code.encode()
        data = (HEAD_REF if cmd.is_ref else HEAD_OP) + code + END

        _LOGGER.debug(
            "Sending %s '%s (%s)'", "ref" if cmd.is_ref else "op", cmd.code, data
        )
        await self._conn.write(data)

        try:
            data = await self._conn.readline()
        except asyncio.TimeoutError:
            _LOGGER.warning("Response timeout for '%s'", cmd.code)
            return

        _LOGGER.debug("Received ack %s", data)

        if not data.startswith(HEAD_ACK + code[0:2]):
            raise JvcProjectorCommandError(
                f"Response ack invalid '{data!r}' for '{cmd.code}' expected '{HEAD_ACK + code[0:2]!r}'"
            )

        if cmd.is_ref:
            try:
                data = await self._conn.readline()
            except asyncio.TimeoutError:
                _LOGGER.warning("Ref response timeout for '%s'", cmd.code)
                return

            _LOGGER.debug("Received ref %s (%s)", data[HEAD_LEN + 2 : -1], data)

            if not data.startswith(HEAD_RES + code[0:2]):
                raise JvcProjectorCommandError(
                    f"Ref ack invalid '{data!r}' for '{cmd.code}'"
                )

            try:
                cmd.response = data[HEAD_LEN + 2 : -1].decode()
            except UnicodeDecodeError:
                cmd.response = data.hex()
                _LOGGER.warning("Failed to decode response '%s'", data)

        cmd.ack = True

    async def disconnect(self) -> None:
        """Disconnect from device."""
        if self._keepalive:
            self._keepalive.cancel()
        await self._disconnect()

    async def _disconnect(self, delay: int = 0) -> None:
        """Disconnect from device."""
        if delay:
            await asyncio.sleep(delay)
        self._keepalive = None
        await self._conn.disconnect()
        _LOGGER.debug("Disconnected")



"""Tests for command module."""

from jvcprojector.command import JvcCommandHelpers


def test_build_command():
    """Test build command succeeds."""
    a = JvcCommandHelpers.get_available_commands()
    assert len(a) > 1



"""Device constants for a JVC Projector."""

from typing import Final

POWER: Final = "power"
OFF: Final = "off"
STANDBY: Final = "standby"
ON: Final = "on"
WARMING: Final = "warming"
COOLING: Final = "cooling"
ERROR: Final = "error"
NORMAL: Final = "normal"
LOW: Final = "low"
MEDIUM: Final = "medium"
HIGH: Final = "high"
AUTO: Final = "auto"

SIGNAL: Final = "signal"
NOSIGNAL: Final = "no_signal"

INPUT: Final = "input"
HDMI1 = "hdmi1"
HDMI2 = "hdmi2"

REMOTE_MENU: Final = "732E"
REMOTE_UP: Final = "7301"
REMOTE_DOWN: Final = "7302"
REMOTE_LEFT: Final = "7336"
REMOTE_RIGHT: Final = "7334"
REMOTE_OK: Final = "732F"
REMOTE_BACK: Final = "7303"
REMOTE_MPC: Final = "73F0"
REMOTE_HIDE: Final = "731D"
REMOTE_INFO: Final = "7374"
REMOTE_INPUT: Final = "7308"
REMOTE_ADVANCED_MENU: Final = "7373"
REMOTE_PICTURE_MODE: Final = "73F4"
REMOTE_COLOR_PROFILE: Final = "7388"
REMOTE_LENS_CONTROL: Final = "7330"
REMOTE_SETTING_MEMORY: Final = "73D4"
REMOTE_GAMMA_SETTINGS: Final = "73F5"
REMOTE_CMD: Final = "738A"
REMOTE_MODE_1: Final = "73D8"
REMOTE_MODE_2: Final = "73D9"
REMOTE_MODE_3: Final = "73DA"
REMOTE_HDMI_1: Final = "7370"
REMOTE_HDMI_2: Final = "7371"
REMOTE_LENS_AP: Final = "7320"
REMOTE_ANAMO: Final = "73C5"
REMOTE_GAMMA: Final = "7375"
REMOTE_COLOR_TEMP: Final = "7376"
REMOTE_3D_FORMAT: Final = "73D6"
REMOTE_PIC_ADJ: Final = "7372"
REMOTE_NATURAL: Final = "736A"
REMOTE_CINEMA: Final = "7368"

# HDR Processing modes
HDR_10PLUS: Final = "hdr10_plus"
HDR_STATIC: Final = "static"
HDR_FRAME_BY_FRAME: Final = "frame_by_frame"
HDR_SCENE_BY_SCENE: Final = "scene_by_scene"

# HDR Content types
HDR_CONTENT_SDR: Final = "sdr"
HDR_CONTENT_HDR10: Final = "hdr10"
HDR_CONTENT_HDR10PLUS: Final = "hdr10_plus"
HDR_CONTENT_HLG: Final = "hlg"
HDR_CONTENT_NONE: Final = "none"

# Anamorphic modes
ANAMORPHIC_A: Final = "a"
ANAMORPHIC_B: Final = "b"
ANAMORPHIC_C: Final = "c"
ANAMORPHIC_D: Final = "d"

# Laser dimming modes
AUTO1: Final = "auto1"
AUTO2: Final = "auto2"
AUTO3: Final = "auto3"

# Aspect Ratio
ASPECT_RATIO_ZOOM: Final = "zoom"
ASPECT_RATIO_NATIVE: Final = "native"

MODEL_MAP = {
    "B5A1": "NZ9",
    "B5A2": "NZ8",
    "B5A3": "NZ7",
    "A2B1": "NX9",
    "A2B2": "NX7",
    "A2B3": "NX5",
    "B2A1": "NX9",
    "B2A2": "NX7",
    "B2A3": "NX5",
    "B5B1": "NP5",
    "XHR1": "X570R",
    "XHR3": "X770R||X970R",
    "XHP1": "X5000",
    "XHP2": "XC6890",
    "XHP3": "X7000||X9000",
    "XHK1": "X500R",
    "XHK2": "RS4910",
    "XHK3": "X700R||X900R",
}

# Key names
KEY_MODEL: Final = "model"
KEY_MAC: Final = "mac"
KEY_VERSION: Final = "version"
KEY_POWER: Final = "power"
KEY_INPUT: Final = "input"
KEY_SOURCE: Final = "source"
KEY_PICTURE_MODE: Final = "picture_mode"
KEY_LOW_LATENCY: Final = "low_latency"
KEY_INSTALLATION_MODE: Final = "installation_mode"
KEY_ANAMORPHIC: Final = "anamorphic"
KEY_HDR: Final = "hdr"
KEY_HDMI_INPUT_LEVEL: Final = "hdmi_input_level"
KEY_HDMI_COLOR_SPACE: Final = "hdmi_color_space"
KEY_COLOR_PROFILE: Final = "color_profile"
KEY_GRAPHICS_MODE: Final = "graphics_mode"
KEY_COLOR_SPACE: Final = "color_space"
KEY_ESHIFT: Final = "eshift"
KEY_LASER_DIMMING: Final = "laser_dimming"
KEY_LASER_VALUE: Final = "laser_value"
KEY_LASER_POWER: Final = "laser_power"
KEY_LASER_TIME: Final = "laser_time"
KEY_MOTION_ENHANCE: Final = "motion_enhance"
KEY_CLEAR_MOTION_DRIVE: Final = "clear_motion_drive"
KEY_HDR_PROCESSING: Final = "hdr_processing"
KEY_HDR_CONTENT_TYPE: Final = "hdr_content_type"
KEY_RESOLUTION: Final = "resolution"


# Command constants
CMD_POWER = "PW"
CMD_INPUT = "IP"
CMD_SOURCE = "SC"
CMD_MODEL = "MD"
CMD_REMOTE = "RC"
CMD_VERSION = "IFSV"
CMD_PICTURE_MODE = "PMPM"
CMD_PICTURE_MODE_INTELLIGENT_APERTURE = "PMDI"
CMD_LASER_DIMMING = "PMDC"
CMD_PICTURE_MODE_COLOR_PROFILE = "PMPR"
CMD_PICTURE_MODE_COLOR_TEMP = "PMCL"
CMD_PICTURE_MODE_COLOR_CORRECTION = "PMCC"
CMD_PICTURE_MODE_GAMMA_TABLE = "PMGT"
CMD_PICTURE_MODE_COLOR_MANAGEMENT = "PMCB"
CMD_PICTURE_MODE_LOW_LATENCY = "PMLL"
CMD_PICTURE_MODE_8K_ESHIFT = "PMUS"
CMD_PICTURE_MODE_CLEAR_MOTION_DRIVE = "PMCM"
CMD_PICTURE_MODE_LASER_VALUE = "PMCV"
CMD_PICTURE_MODE_MOTION_ENHANCE = "PMME"
CMD_PICTURE_MODE_LASER_POWER = "PMLP"
CMD_PICTURE_MODE_GRAPHICS_MODE = "PMGM"
CMD_INPUT_SIGNAL_HDMI_INPUT_LEVEL = "ISIL"
CMD_INPUT_SIGNAL_HDMI_COLOR_SPACE = "ISHS"
CMD_INPUT_SIGNAL_HDMI_2D_3D = "IS3D"
CMD_INPUT_SIGNAL_ASPECT = "ISAS"
CMD_INPUT_SIGNAL_MASK = "ISMA"
CMD_INSTALLATION_MODE = "INML"
CMD_INSTALLATION_LENS_CONTROL = "INLC"
CMD_INSTALLATION_LENS_IMAGE_PATTERN = "INIP"
CMD_INSTALLATION_LENS_LOCK = "INLL"
CMD_INSTALLATION_SCREEN_ADJUST = "INSC"
CMD_INSTALLATION_STYLE = "INIS"
CMD_INSTALLATION_ANAMORPHIC = "INVS"
CMD_DISPLAY_BACK_COLOR = "DSBC"
CMD_DISPLAY_MENU_POSITION = "DSMP"
CMD_DISPLAY_SOURCE_DISPLAY = "DSSD"
CMD_DISPLAY_LOGO = "DSLO"
CMD_FUNCTION_TRIGGER = "FUTR"
CMD_FUNCTION_OFF_TIMER = "FUOT"
CMD_FUNCTION_ECO_MODE = "FUEM"
CMD_FUNCTION_CONTROL4 = "FUCF"
CMD_FUNCTION_INPUT = "IFIN"
CMD_FUNCTION_SOURCE = "IFIS"
CMD_FUNCTION_DEEP_COLOR = "IFDC"
CMD_FUNCTION_COLOR_SPACE = "IFXV"
CMD_FUNCTION_COLORIMETRY = "IFCM"
CMD_FUNCTION_HDR = "IFHR"
CMD_FUNCTION_LASER_TIME = "IFLT"
CMD_PICTURE_MODE_HDR_LEVEL = "PMHL"
CMD_PICTURE_MODE_HDR_PROCESSING = "PMHP"
CMD_PICTURE_MODE_HDR_CONTENT_TYPE = "PMCT"
CMD_PICTURE_MODE_THEATER_OPTIMIZER = "PMNM"
CMD_PICTURE_MODE_THEATER_OPTIMIZER_LEVEL = "PMNL"
CMD_PICTURE_MODE_THEATER_OPTIMIZER_PROCESSING = "PMNP"
CMD_LAN_SETUP_DHCP = "LSDS"
CMD_LAN_SETUP_MAC_ADDRESS = "LSMA"
CMD_LAN_SETUP_IP_ADDRESS = "LSIP"

# Map keys to commands to make sending commands simple
KEY_MAP_TO_COMMAND: Final = {
    KEY_POWER: CMD_POWER,
    KEY_INPUT: CMD_INPUT,
    KEY_SOURCE: CMD_SOURCE,
    KEY_PICTURE_MODE: CMD_PICTURE_MODE,
    KEY_INSTALLATION_MODE: CMD_INSTALLATION_MODE,
    KEY_ANAMORPHIC: CMD_INSTALLATION_ANAMORPHIC,
    KEY_HDR: CMD_FUNCTION_HDR,
    KEY_HDMI_INPUT_LEVEL: CMD_INPUT_SIGNAL_HDMI_INPUT_LEVEL,
    KEY_HDMI_COLOR_SPACE: CMD_INPUT_SIGNAL_HDMI_COLOR_SPACE,
    KEY_COLOR_PROFILE: CMD_PICTURE_MODE_COLOR_PROFILE,
    KEY_GRAPHICS_MODE: CMD_PICTURE_MODE_GRAPHICS_MODE,
    KEY_COLOR_SPACE: CMD_FUNCTION_COLOR_SPACE,
    KEY_ESHIFT: CMD_PICTURE_MODE_8K_ESHIFT,
    KEY_LASER_DIMMING: CMD_LASER_DIMMING,
    KEY_LASER_VALUE: CMD_PICTURE_MODE_LASER_VALUE,
    KEY_LASER_POWER: CMD_PICTURE_MODE_LASER_POWER,
    KEY_LASER_TIME: CMD_FUNCTION_LASER_TIME,
    KEY_MOTION_ENHANCE: CMD_PICTURE_MODE_MOTION_ENHANCE,
    KEY_CLEAR_MOTION_DRIVE: CMD_PICTURE_MODE_CLEAR_MOTION_DRIVE,
    KEY_HDR_PROCESSING: CMD_PICTURE_MODE_HDR_PROCESSING,
    KEY_HDR_CONTENT_TYPE: CMD_PICTURE_MODE_HDR_CONTENT_TYPE,
}

# Value constants
VAL_POWER = [STANDBY, ON, COOLING, WARMING, ERROR]
VAL_INPUT = {
    "0": "svideo",
    "1": "video",
    "2": "component",
    "3": "pc",
    "6": HDMI1,
    "7": HDMI2,
}
VAL_SOURCE = [NOSIGNAL, SIGNAL]
VAL_PICTURE_MODE = {
    "00": "film",
    "01": "cinema",
    "02": "animation",
    "03": "natural",
    "04": "hdr10",
    "06": "thx",
    "0B": "frameadapt_hdr",
    "0C": "user1",
    "0D": "user2",
    "0E": "user3",
    "0F": "user4",
    "10": "user5",
    "11": "user6",
    "14": "hlg",
    "15": "hdr10+",
    "16": "pana_pq",
    "17": "filmmaker",
    "18": "frameadapt_hdr2",
    "19": "frameadapt_hdr3",
}
VAL_INTELLIGENT_APERTURE = [OFF, AUTO1, AUTO2]
VAL_LASER_DIMMING = [OFF, AUTO1, AUTO2, AUTO3]
VAL_COLOR_PROFILE = {
    "00": OFF,
    "01": "film1",
    "02": "film2",
    "03": "bt709",
    "04": "cinema",
    "05": "cinema2",
    "06": "anime",
    "07": "anime2",
    "08": "video",
    "09": "vivid",
    "0A": "hdr",
    "0B": "bt2020(wide)",
    "0C": "3d",
    "0D": "thx",
    "0E": "custom1",
    "0F": "custom2",
    "10": "custom3",
    "11": "custom4",
    "12": "custom5",
    "21": "dci",
    "22": "custom6",
    "24": "bt2020(normal)",
    "25": "off(wide)",
    "26": AUTO,
}
VAL_COLOR_TEMP = {
    "00": "5500k",
    "02": "6500k",
    "04": "7500k",
    "08": "9300k",
    "09": "high",
    "0A": "custom1",
    "0B": "custom2",
    "0C": "hdr10",
    "0D": "xenon1",
    "0E": "xenon2",
    "14": "hlg",
}
VAL_COLOR_CORRECTION = {
    "00": "5500k",
    "02": "6500k",
    "04": "7500k",
    "08": "9300k",
    "09": "high",
    "0D": "xenon1",
    "0E": "xenon2",
}
VAL_GAMMA_TABLE = {
    "00": "2.2",
    "01": "cinema1",
    "02": "cinema2",
    "04": "custom1",
    "05": "custom2",
    "06": "custom3",
    "07": "hdr_hlg",
    "08": "2.4",
    "09": "2.6",
    "0A": "film1",
    "0B": "film2",
    "0C": "hdr_pq",
    "0D": "pana_pq",
    "10": "thx",
    "15": "hdr_auto",
}
VAL_TOGGLE = [OFF, ON]
VAL_CLEAR_MOTION_DRIVE = [OFF, "none", LOW, HIGH, "inverse_telecine"]
VAL_MOTION_ENHANCE = [OFF, LOW, HIGH]
# note this order can't change due to how building comamands works. JVC historically did not have medium it was added with the NZ series. It is 2 for medium, 1 for high.
VAL_LASER_POWER = [
    LOW,
    HIGH,
    MEDIUM,
]
VAL_GRAPHICS_MODE = ["standard", "high-res", "high-res2", OFF]
VAL_HDMI_INPUT_LEVEL = ["standard", "enhanced", "super_white", AUTO]
VAL_HDMI_COLOR_SPACE = [AUTO, "ycbcr(4:4:4)", "ycbcr(4:2:2)", "rgb"]
VAL_HDMI_2D_3D = ["2d", AUTO, "none", "side_by_side", "top_bottom"]
VAL_ASPECT = ["none", "zoom", AUTO, "native"]
VAL_MASK = ["none", ON, OFF]
VAL_INSTALLATION_MODE = [f"mode{i}" for i in range(1, 11)]
VAL_LENS_CONTROL = ["stop", "start"]
VAL_INSTALLATION_STYLE = ["front", "front_ceiling", "rear", "rear_ceiling"]
VAL_ANAMORPHIC = [OFF, ANAMORPHIC_A, ANAMORPHIC_B, ANAMORPHIC_C, ANAMORPHIC_D]
VAL_BACK_COLOR = ["blue", "black"]
VAL_MENU_POSITION = [
    "left-top",
    "right-top",
    "center",
    "left-bottom",
    "right-bottom",
    "left",
    "right",
]
VAL_TRIGGER = [OFF, "power", "anamo"] + [f"ins{i}" for i in range(1, 11)]
VAL_OFF_TIMER = [OFF, "1hour", "2hour", "3hour", "4hour"]
VAL_FUNCTION_INPUT = [HDMI1, HDMI2]
VAL_FUNCTION_SOURCE = {
    "02": "480p",
    "03": "576p",
    "04": "720p50",
    "05": "720p60",
    "06": "1080i50",
    "07": "1080i60",
    "08": "1080p24",
    "09": "1080p50",
    "0A": "1080p60",
    "0B": "nosignal",
    "0C": "720p3d",
    "0D": "1080i3d",
    "0E": "1080p3d",
    "0F": "outofrange",
    "10": "4k(4096)60",
    "11": "4k(4096)50",
    "12": "4k(4096)30",
    "13": "4k(4096)25",
    "14": "4k(4096)24",
    "15": "4k(3840)60",
    "16": "4k(3840)50",
    "17": "4k(3840)30",
    "18": "4k(3840)25",
    "19": "4k(3840)24",
    "1C": "1080p25",
    "1D": "1080p30",
    "1E": "2048x1080p24",
    "1F": "2048x1080p25",
    "20": "2048x1080p30",
    "21": "2048x1080p50",
    "22": "2048x1080p60",
    "23": "3840x2160p120",
    "24": "4096x2160p120",
    "25": "vga(640x480)",
    "26": "svga(800x600)",
    "27": "xga(1024x768)",
    "28": "sxga(1280x1024)",
    "29": "wxga(1280x768)",
    "2A": "wxga+(1440x900)",
    "2B": "wsxga+(1680x1050)",
    "2C": "wuxga(1920x1200)",
    "2D": "wxga(1280x800)",
    "2E": "fwxga(1366x768)",
    "2F": "wxga++(1600x900)",
    "30": "uxga(1600x1200)",
    "31": "qxga",
    "32": "wqxga",
}
VAL_DEEP_COLOR = ["8bit", "10bit", "12bit"]
VAL_COLOR_SPACE = ["rgb", "yuv"]
VAL_COLORIMETRY = [
    "nodata",
    "bt601",
    "bt709",
    "xvycc601",
    "xvycc709",
    "sycc601",
    "adobe_ycc601",
    "adobe_rgb",
    "bt2020(constant_luminance)",
    "bt2020(non-constant_luminance)",
    "srgb",
]
VAL_HDR_MODES = [
    HDR_CONTENT_SDR,
    KEY_HDR,
    "smpte_st_2084",
    "hybrid_log",
    "none",
]
VAL_HDR = {
    "0": VAL_HDR_MODES[0],
    "1": VAL_HDR_MODES[1],
    "2": VAL_HDR_MODES[2],
    "3": VAL_HDR_MODES[3],
    "F": VAL_HDR_MODES[4],
}
VAL_HDR_LEVEL = [AUTO, "-2", "-1", "0", "1", "2"]
VAL_HDR_PROCESSING = ["static", "frame", "scene"]
VAL_HDR_CONTENT_TYPE = [
    AUTO,
    HDR_CONTENT_SDR,
    "none",
    HDR_CONTENT_HDR10,
    HDR_CONTENT_HLG,
]
VAL_THEATER_OPTIMIZER_LEVEL = ["reserved", LOW, MEDIUM, HIGH]
VAL_THEATER_OPTIMIZER_PROCESSING = ["-", "start"]
